"""
The TMP102 is a two-wire, serial output temperature
sensor available in a tiny SOT563 package. Requiring
no external components, the TMP102 is capable of
reading temperatures to a resolution of 0.0625°C.

https://www.sparkfun.com/datasheets/Sensors/Temperature/tmp102.pdf

"""
